/*----------------------

   GATE - Geant4 Application for Tomographic Emission 
   OpenGATE Collaboration 
     
   Giovanni Santin <giovanni.santin@cern.ch>
   Daniel Strul <daniel.strul@iphe.unil.ch> 
   Claude Comtat <comtat@ieee.org>
     
   Copyright (C) 2002,2003 UNIL/IPHE, CH-1015 Lausanne 
   Copyright (C) 2003 CEA/SHFJ, F-91401 Orsay 

This software is distributed under the terms 
of the GNU Lesser General  Public Licence (LGPL) 
See GATE/LICENSE.txt for further details 
----------------------*/

#include "GateRunManager.hh"
#include <queue>
#include "G4UImanager.hh"
#include "G4UIterminal.hh"
#include "G4UItcsh.hh"
#ifdef G4UI_USE_ROOT
#include "G4UIRoot.hh"
#endif
#include "G4ScoringManager.hh"

#include "GateDetectorConstruction.hh"
#include "GatePhysicsList.hh"
#include "GatePrimaryGeneratorAction.hh"
#include "GateRunAction.hh"
#include "GateEventAction.hh"
#include "GateTrackingAction.hh"
#include "GateSteppingAction.hh"
#include "GateStackingAction.hh"

#include "GateApplicationMgr.hh"
#include "GateSourceMgr.hh"
#include "GateSignalHandler.hh"
#include "GateClock.hh"
#include "GateOutputMgr.hh"
#include "GateUIcontrolMessenger.hh"
#include "GateUIterminal.hh"
#include "GateDigitizer.hh"
#include "GatePulseProcessorChain.hh"

#ifdef G4VIS_USE
#include "G4VisExecutive.hh"
#endif

#ifdef G4ANALYSIS_USE_ROOT
#include "GateHitFileReader.hh"
#endif

#ifdef G4VMC
#include "TG4XMLGeometryGenerator.h"
#include "TG4XMLMessenger.h"
#endif

#ifndef G4ANALYSIS_USE_ROOT
void AbortIfRootNotFound()
{
    G4cout  << G4endl 
      	    << "Sorry, but it seems that GATE was compiled without the ROOT option." << G4endl
      	    << "Consequently, you can not run GATE in DigiGate mode, and the execution will abort." << G4endl
	    << G4endl
	    << "There maybe several reasons why GATE was compiled without the ROOT option:" << G4endl
	    << G4endl
	    << "1) ROOT is not installed on your system;" << G4endl
	    << "2) You did not source a GATE configuration script before compiling GATE;" << G4endl
	    << "3) The configuration script you used has been modified to disable the ROOT option;" << G4endl
	    << "4) You used the configuration file 'env_gate.csh' but it did not set the ROOT option." << G4endl
	    << G4endl
      	    << "Here is what you can do:" << G4endl
	    << G4endl
	    << "1) I'm sorry but you won't have access to DigiGate, as it needs ROOT to work." << G4endl 
	    << "   We apologize for this inconvenience, but there is nothing we can do about it," << G4endl
	    << "   because DigiGate works by re-reading a ROOT hit-file" << G4endl 
	    << G4endl
	    << "2,3) You must enable the ROOT option with:" << G4endl
	    << "         setenv G4ANALYSIS_USE_ROOT 1 (csh)" << G4endl
	    << "         export G4ANALYSIS_USE_ROOT=1 (sh)"  << G4endl
	    << "   Then rebuild the ROOT-dependent part of GATE (see below)" << G4endl 
	    << G4endl
	    << "4) The script 'env_gate.csh' decides whether or not to set the ROOT option by checking the value of " << G4endl
	    << "   the environment variable 'ROOTSYS'. Maybe ROOT is installed on your system but ROOTSYS is not set." << G4endl
	    << "   If that's the case, you must set the variable ROOTSYS with:" << G4endl
	    << "         setenv ROOTSYS the_home_dir_of_ROOT (csh)" << G4endl
	    << "         export ROOTSYS=the_home_dir_of_ROOT (sh)"  << G4endl
	    << "   Then source 'env_gate.csh' again, then rebuild the ROOT-dependent part of GATE (see below)" << G4endl 
	    << G4endl
	    << "Note: to rebuild the ROOT-dependent part of GATE, use the following command:" << G4endl
	    << "         touch Gate.cc `grep -l G4ANALYSIS_USE_ROOT include/* src/*`" << G4endl
	    << "         make" << G4endl
	    << G4endl;
    G4Exception("Correct problem then try again... Sorry!");	    
}
#endif


// Variables decoded from the Gate arguments
DigiMode aDigiMode=kruntimeMode; 
G4bool isBatchMode = false;
char* batchFileName = 0;
std::queue<G4String> commandQueue;
// Decode the arguments passed to Gate
void DecodeArguments(int argc,char** argv)
{
  int nextArg = 1;

  // Loop while there are optional arguments on the command line
  while (nextArg<argc)
  {
    if (!strcmp(argv[nextArg],"-d")) {
      // Select DigiGate mode
      aDigiMode=kofflineMode;
      nextArg++;
    }
    else if (!strcmp(argv[nextArg],"-a")) {
      // Pass an alias to Gate
      if ( argc <= (nextArg+2) )
        G4Exception("Not enough arguments after the option flag '-a'\n"
	            "The syntax to use this flag, which allows to create an alias, is:\n"
		    "\tGate -a ALIAS_NAME ALIAS_VALUE\n\n"
		    "Aborting!");
	//G4String newCommand = G4String("/control/alias ") + argv[++nextArg] + " \"" + argv[++nextArg] + "\"";
	nextArg++ ; 
	G4String newCommand = G4String("/control/alias ") + argv[nextArg] ;
        nextArg++ ; 
	newCommand = newCommand + " \"" + argv[nextArg] + "\"";
	commandQueue.push(newCommand);	    
        nextArg++;
    }
#ifdef G4ANALYSIS_USE_ROOT
    else if (!strcmp(argv[nextArg],"-b")) {
      // disable X connection if asked...
      gROOT->SetBatch();
      nextArg++;
    }
#endif
    else {
      // The argument was not recognised: exit
      break;
    }
  }

  // If there is an argument left on the command line, select the batch mode
  // and use it as the batch-file name
  isBatchMode = (argc!=nextArg);
  if (isBatchMode)
  	batchFileName = argv[nextArg];
}




// Execute the list of commands obtained from the command-line arguments
void ExecuteCommandQueue()
{
  G4UImanager* anUImanager = G4UImanager::GetUIpointer();

  while ( commandQueue.size() )
  { 
    G4cout << commandQueue.front() << G4endl;
    anUImanager->ApplyCommand(commandQueue.front());
    commandQueue.pop() ; 
  }
}


int main(int argc,char** argv)
{
  gROOT->GetPluginManager()->AddHandler("TVirtualStreamerInfo", "*", "TStreamerInfo", "RIO", 
"TStreamerInfo()");

//
// random engine
//

  //CLHEP::HepJamesRandom randomEngine;
  //CLHEP::Ranlux64Engine randomEngine;
    CLHEP::MTwistEngine randomEngine;
    CLHEP::HepRandom::setTheEngine(&randomEngine);

  
  // Call the srugement decoding function
  DecodeArguments(argc,argv);

  // Install the signal handler to handle interrupt calls
  GateSignalHandler::Install();

  // Create the mandatory class collaboration
  GateRunManager* runManager = new GateRunManager;
  //LF
  //G4ScoringManager * scManager = G4ScoringManager::GetScoringManager();
  //scManager->SetVerboseLevel(0);
//LF
  GateDetectorConstruction* gateDC = new GateDetectorConstruction;
  runManager->SetUserInitialization(gateDC);
  runManager->SetUserInitialization(new GatePhysicsList);
#ifdef G4VIS_USE
    // visualization manager
    G4VisManager* visManager = new G4VisExecutive;
    visManager->Initialize();
#endif

  // Initialise the geometry 
  runManager->InitGeometryOnly();
  // Incorporate the optional user actions
  runManager->SetUserAction(new GatePrimaryGeneratorAction());
  runManager->SetUserAction(new GateRunAction);
  runManager->SetUserAction(new GateEventAction);
  runManager->SetUserAction(new GateTrackingAction);
  runManager->SetUserAction(new GateSteppingAction);
  runManager->SetUserAction(new GateStackingAction);

  // Create various singleton objets
#ifdef G4ANALYSIS_USE
  GateOutputMgr::SetDigiMode(aDigiMode);
  GateOutputMgr* outputMgr = GateOutputMgr::GetInstance();
  GateDigitizer* digitizer = GateDigitizer::GetInstance();
  GatePulseProcessorChain* singleChain = new GatePulseProcessorChain(digitizer,"Singles"); 
  digitizer->StoreNewPulseProcessorChain(singleChain);
#endif

  if (aDigiMode==kofflineMode)
#ifdef G4ANALYSIS_USE_ROOT
    GateHitFileReader::GetInstance();
#else
    AbortIfRootNotFound();
#endif
  GateSourceMgr*      sourceMgr = GateSourceMgr::GetInstance();
  GateApplicationMgr* appMgr    = GateApplicationMgr::GetInstance();
  GateClock::GetInstance()->SetTime(0);
  /*GateUIcontrolMessenger* controlMessenger = */new GateUIcontrolMessenger;
#ifdef G4VMC
  new TG4XMLMessenger(new TG4XMLGeometryGenerator);
#endif

  // Start the execution of GATE
  // Two modes are possible: batch or interactive
  if(!isBatchMode)
  {
G4cout<<""<<G4endl;
G4cout<<""<<G4endl;
G4cout<<"**********************************************************************"<<G4endl;
G4cout<<" GATE version name: gate_v4.0.0  (20-August-2008)                     "<<G4endl;
G4cout<<"                    Copyright : OpenGATE Collaboration                "<<G4endl;
G4cout<<"                    Reference : Phys. Med. Biol. 49 (2004) 4543-4561  "<<G4endl;
G4cout<<"                    WWW : http://opengatecollaboration.healthgrid.org/"<<G4endl;
G4cout<<"**********************************************************************"<<G4endl;
G4cout<<""<<G4endl;

    // Start Gate in interactive mode
    
    G4UIsession *session = (G4UIsession *)0;
    G4UItcsh* tcsh = new G4UItcsh;

#ifdef G4UI_USE_ROOT
    // G4UIRoot is a ROOT based GUI.
    G4cout << "Creating the ROOT UI session..." << G4endl;
    session = new G4UIRoot(argc,argv,"GATE", "Geant4 Application for Tomographic Emission","root_logo.xpm", "logoGate_medium.xpm");
#else
    session = new GateUIterminal(tcsh);      
#endif

    // Execute the list of commands obtained from the command-line arguments
    ExecuteCommandQueue();

    G4UImanager::GetUIpointer()->ApplyCommand("/gate/control/execute prerunGate.mac");
    session->SessionStart();
    delete session;



  } else {
    // Start Gate in batch mode

    // Execute the list of commands obtained from the command-line arguments
    ExecuteCommandQueue();

    G4String command = "/gate/control/execute ";
    G4UImanager::GetUIpointer()->ApplyCommand( command + batchFileName );

  }

#ifdef G4ANALYSIS_USE
  if (outputMgr) delete outputMgr;
#endif

#ifdef G4VIS_USE
  delete visManager;
#endif

  delete sourceMgr;
  delete appMgr;
  delete runManager;

  return 0;
}
